@@ -1,4 +1,7 @@
 /*
+ * NOTE: this is part of libzzipmmapped (i.e. it is not libzzip).
+ *                                            ==================
+ *
  * These routines are fully independent from the traditional zzip
  * implementation. They assume a readonly mmapped sharedmem block
  * representing a complete zip file. The functions show how to 
@@ -14,7 +17,7 @@
  * Author: 
  *      Guido Draheim <guidod@gmx.de>
  *
- * Copyright (c) 2003,2004 Guido Draheim
+ * Copyright (c) 2003,2004,2006 Guido Draheim
  *          All rights reserved,
  *          use under the restrictions of the 
  *          Lesser GNU General Public License
@@ -22,14 +25,21 @@
  *          of the Mozilla Public License 1.1
  */
 
+#define _ZZIP_DISK_FILE_STRUCT 1
+
+#ifdef __linux__
+#define _GNU_SOURCE _glibc_developers_are_idiots_to_call_strndup_gnu_specific_
+#endif
+
 #include <zzip/mmapped.h>
+#include <zzip/format.h>
+#include <zzip/fetch.h>
+#include <zzip/__mmap.h>
+#include <zzip/__fnmatch.h>
+
 #include <stdlib.h>
 #include <sys/stat.h>
 
-#ifdef ZZIP_HAVE_FNMATCH_H
-#include <fnmatch.h>
-#endif
-
 #if   defined ZZIP_HAVE_UNISTD_H
 #include <unistd.h>
 #elif defined ZZIP_HAVE_IO_H
@@ -42,10 +52,6 @@
 #include <strings.h>
 #endif
 
-#include <zlib.h>
-#include <zzip/format.h>
-#include <zzip/fetch.h>
-#include <zzip/__mmap.h>
 
 #if __STDC_VERSION__+0 > 199900L
 #define ___
@@ -59,10 +65,10 @@
  * This function does primary initialization of a disk-buffer struct.
  */
 int
-zzip_disk_init(ZZIP_DISK* disk, char* buffer, zzip_size_t buflen)
+zzip_disk_init(ZZIP_DISK* disk, void* buffer, zzip_size_t buflen)
 {
-    disk->buffer = buffer;
-    disk->endbuf = buffer+buflen;
+    disk->buffer = (zzip_byte_t*) buffer;
+    disk->endbuf = (zzip_byte_t*) buffer + buflen;
     disk->reserved = 0;
     disk->flags = 0;
     disk->mapped = 0;
@@ -74,7 +80,7 @@ zzip_disk_init(ZZIP_DISK* disk, char* buffer, zzip_size_t buflen)
 /** => zzip_disk_mmap
  * This function allocates a new disk-buffer with => malloc(3)
  */
-ZZIP_DISK* _zzip_restrict
+zzip__new__ ZZIP_DISK*
 zzip_disk_new(void)
 {
     ZZIP_DISK* disk = malloc(sizeof(disk));
@@ -90,13 +96,13 @@ zzip_disk_new(void)
  * successful then a newly allocated ZZIP_DISK* is returned with 
  * disk->buffer pointing to the mapview of the zipdisk content.
  */
-ZZIP_DISK* _zzip_restrict
+zzip__new__ ZZIP_DISK*
 zzip_disk_mmap(int fd)
 {
     struct stat st;
-    if (fstat (fd, &st) || !st.st_size) return 0;
+    if (fstat (fd, &st) || ! st.st_size) return 0;
     ___ ZZIP_DISK* disk = zzip_disk_new (); if (! disk) return 0;
-    disk->buffer = _zzip_mmap (& zzip->mapped, fd, 0, st.st_size);
+    disk->buffer = _zzip_mmap (& disk->mapped, fd, 0, st.st_size);
     if (disk->buffer == MAP_FAILED) { free (disk); return 0; }
     disk->endbuf = disk->buffer + st.st_size;
     return disk; ____;
@@ -123,19 +129,19 @@ zzip_disk_munmap(ZZIP_DISK* disk)
  * memory block. Only if that fails too then we return null. Since handling
  * of disk->buffer is ambigous it should not be snatched away please.
  */
-ZZIP_DISK* _zzip_restrict
+ZZIP_DISK* zzip__new__
 zzip_disk_open(char* filename)
 {
 #  ifndef O_BINARY
 #  define O_BINARY 0
 #  endif
     struct stat st;
-    if (stat (filename, &st) || !st.st_size) return 0;
+    if (stat (filename, &st) || ! st.st_size) return 0;
     ___ int fd = open (filename, O_RDONLY|O_BINARY);
     if (fd <= 0) return 0;
     ___ ZZIP_DISK* disk = zzip_disk_mmap (fd);
     if (disk) return disk;
-    ___ char* buffer = malloc (st.st_size);
+    ___ zzip_byte_t* buffer = malloc (st.st_size);
     if (! buffer) return 0;
     if ((st.st_size == read (fd, buffer, st.st_size)) &&
 	(disk = zzip_disk_new ())) 
@@ -170,10 +176,10 @@ zzip_disk_close(ZZIP_DISK* disk)
 #define _zzip_strndup strndup
 #else
 /* if your system does not have strndup: */
-static char* _zzip_restrict _zzip_strndup(char* p, int maxlen)
+zzip__new__ static char* _zzip_strndup(char* p, size_t maxlen)
 {
     if (! p) return 0;
-    ___ char* r = malloc (maxlen+1);
+    ___ zzip_byte_t* r = malloc (maxlen+1);
     if (! r) return r;
     strncpy (r, p, maxlen);
     r[maxlen] = '\0';
@@ -207,7 +213,7 @@ static int _zzip_strcasecmp(char* __zzip_restrict a, char* _zzip_restrict b)
  * the data block right after the file_header. Only disk->buffer would be
  * needed to perform the seek but we check the mmapped range end as well.
  */
-char*
+zzip_byte_t*
 zzip_disk_entry_to_data(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
 {
     struct zzip_file_header* file = 
@@ -223,7 +229,7 @@ zzip_disk_entry_to_data(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
 struct zzip_file_header*
 zzip_disk_entry_to_file_header(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
 {
-    char* file_header = /* (struct zzip_file_header*) */
+    zzip_byte_t* file_header = /* (struct zzip_file_header*) */
 	(disk->buffer + zzip_disk_entry_fileoffset (entry));
     if (disk->buffer > file_header || file_header >= disk->endbuf) 
 	return 0;
@@ -237,7 +243,7 @@ zzip_disk_entry_to_file_header(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
  * in the zip central directory but if not then we fallback to the filename
  * given in the file_header of each compressed data portion.
  */
-char* _zzip_restrict
+zzip__new__ char*
 zzip_disk_entry_strdup_name(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
 {
     if (! disk || ! entry) return 0;
@@ -252,12 +258,35 @@ zzip_disk_entry_strdup_name(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
     else
 	return 0;
 
-    if (disk->buffer > name || name+len > disk->endbuf)
+    if ((zzip_byte_t*) name < disk->buffer || 
+	(zzip_byte_t*) name+len > disk->endbuf)
 	return 0;
     
     return  _zzip_strndup (name, len); ____;
 }
 
+/** => zzip_disk_entry_to_data
+ * This function is similar creating a reference to a zero terminated
+ * string but it can only exist in the zip central directory entry.
+ */
+zzip__new__ char*
+zzip_disk_entry_strdup_comment(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
+{
+    if (! disk || ! entry) return 0;
+
+    ___ char* text; zzip_size_t len;
+    if ((len = zzip_disk_entry_comment (entry)))
+	text = zzip_disk_entry_to_comment (entry);
+    else
+	return 0;
+
+    if ((zzip_byte_t*) text < disk->buffer || 
+	(zzip_byte_t*) text+len > disk->endbuf)
+	return 0;
+    
+    return  _zzip_strndup (text, len); ____;
+}
+
 /* ====================================================================== */
 
 /** => zzip_disk_findfile
@@ -292,25 +321,32 @@ zzip_disk_findfirst(ZZIP_DISK* disk)
 {
     if (disk->buffer > disk->endbuf-sizeof(struct zzip_disk_trailer))
 	return 0;
-    ___ char* p = disk->endbuf-sizeof(struct zzip_disk_trailer);
+    ___ zzip_byte_t* p = disk->endbuf-sizeof(struct zzip_disk_trailer);
     for (; p >= disk->buffer ; p--)
     {
-	if (! zzip_disk_trailer_check_magic(p)) continue;
-	___ char* root = /* (struct zzip_disk_entry*) */ disk->buffer + 
-	    zzip_disk_trailer_get_rootseek ((struct zzip_disk_trailer*)p);
-	if (root > p) 
-	{   /* the first disk_entry is after the disk_trailer? can't be! */
-	    zzip_size_t rootsize =
-		zzip_disk_trailer_get_rootsize ((struct zzip_disk_trailer*)p);
-	    if (disk->buffer+rootsize > p) continue;
-	    /* a common brokeness that can be fixed: we just assume that the 
-	     * central directory was written directly before the trailer: */
-	    root = p - rootsize;
-	}
+	zzip_byte_t* root; /* (struct zzip_disk_entry*) */
+	if (zzip_disk_trailer_check_magic(p)) {
+	    root =  disk->buffer + zzip_disk_trailer_get_rootseek (
+		(struct zzip_disk_trailer*)p);
+	    if (root > p) 
+	    {   /* the first disk_entry is after the disk_trailer? can't be! */
+		zzip_size_t rootsize = zzip_disk_trailer_get_rootsize (
+		    (struct zzip_disk_trailer*)p);
+		if (disk->buffer+rootsize > p) continue;
+		/* a common brokeness that can be fixed: we just assume the
+		 * central directory was written directly before the trailer:*/
+		root = p - rootsize;
+	    }
+	} else if (zzip_disk64_trailer_check_magic(p)) {
+	    if (sizeof(void*) < 8) return 0; /* EOVERFLOW */
+	    root =  disk->buffer + zzip_disk64_trailer_get_rootseek (
+		(struct zzip_disk64_trailer*)p);
+	    if (root > p) continue; 
+	} else continue;
+
 	if (root < disk->buffer) continue;
 	if (zzip_disk_entry_check_magic(root)) 
 	    return (struct zzip_disk_entry*) root;
-	____;
     }____;
     return 0;
 }
@@ -324,12 +360,14 @@ zzip_disk_findfirst(ZZIP_DISK* disk)
 struct zzip_disk_entry*
 zzip_disk_findnext(ZZIP_DISK* disk, struct zzip_disk_entry* entry)
 {
-    if ((char*)entry < disk->buffer || 
-	(char*)entry > disk->endbuf-sizeof(entry) ||
+    if ((zzip_byte_t*)entry < disk->buffer || 
+	(zzip_byte_t*)entry > disk->endbuf-sizeof(entry) ||
+	! zzip_disk_entry_check_magic (entry) ||
 	zzip_disk_entry_sizeto_end (entry) > 64*1024)
 	return 0;
     entry = zzip_disk_entry_to_next_entry (entry);
-    if ((char*)entry > disk->endbuf-sizeof(entry) ||
+    if ((zzip_byte_t*)entry > disk->endbuf-sizeof(entry) ||
+	! zzip_disk_entry_check_magic (entry) ||
 	zzip_disk_entry_sizeto_end (entry) > 64*1024 ||
 	zzip_disk_entry_skipto_end (entry) + sizeof(entry) > disk->endbuf)
 	return 0;
@@ -370,23 +408,6 @@ zzip_disk_findfile(ZZIP_DISK* disk, char* filename,
     return 0;
 }
 
-#ifdef ZZIP_HAVE_FNMATCH_H
-#define _zzip_fnmatch fnmatch
-# ifdef FNM_CASEFOLD
-# define _zzip_fnmatch_CASEFOLD FNM_CASEFOLD
-# else
-# define _zzip_fnmatch_CASEFOLD 0
-# endif
-#else
-# define _zzip_fnmatch_CASEFOLD 0
-/* if your system does not have fnmatch, we fall back to strcmp: */
-static int _zzip_fnmatch(char* pattern, char* string, int flags)
-{ 
-    puts ("<zzip:mmapped:strcmp>");
-    return strcmp (pattern, string); 
-}
-#endif
-
 /** => zzip_disk_findfile
  *
  * This function uses a compare-function with an additional argument
@@ -427,19 +448,6 @@ zzip_disk_findmatch(ZZIP_DISK* disk, char* filespec,
 
 /* ====================================================================== */
 
-/**
- * typedef struct zzip_disk_file ZZIP_DISK_FILE;
- */
-struct zzip_disk_file
-{
-    char* buffer;                      /* fopen disk->buffer */
-    char* endbuf;                      /* fopen disk->endbuf */
-    struct zzip_file_header* header;   /* fopen detected header */
-    zzip_size_t avail;                 /* memorized for checks on EOF */
-    z_stream zlib;                     /* for inflated blocks */
-    char* stored;                      /* for stored blocks */
-};
-
 /** => zzip_disk_fopen
  *
  * the ZZIP_DISK_FILE* is rather simple in just encapsulating the
@@ -451,32 +459,34 @@ struct zzip_disk_file
  * _read() operations will be able to get the next data portion or
  * return an eof condition for that file part wrapped in the zip archive.
  */
-ZZIP_DISK_FILE* _zzip_restrict
+zzip__new__ ZZIP_DISK_FILE*
 zzip_disk_entry_fopen (ZZIP_DISK* disk, ZZIP_DISK_ENTRY* entry)
 {
-    ZZIP_DISK_FILE* file = malloc(sizeof(ZZIP_DISK_FILE));
+    /* keep this in sync with zzip_mem_entry_fopen */
+    struct zzip_file_header* header = 
+	zzip_disk_entry_to_file_header (disk, entry);
+    if (! header) return 0;
+    ___ ZZIP_DISK_FILE* file = malloc(sizeof(ZZIP_DISK_FILE));
     if (! file) return file;
     file->buffer = disk->buffer;
     file->endbuf = disk->endbuf;
-    file->header = zzip_disk_entry_to_file_header (disk, entry);
-    if (! file->header) { free (file); return 0; }
-    file->avail = zzip_file_header_usize (file->header);
+    file->avail = zzip_file_header_usize (header);
 
-    if (! file->avail || zzip_file_header_data_stored (file->header))
-    { file->stored = zzip_file_header_to_data (file->header); return file; }
+    if (! file->avail || zzip_file_header_data_stored (header))
+    { file->stored = zzip_file_header_to_data (header); return file; }
 
     file->stored = 0;
     file->zlib.opaque = 0;
     file->zlib.zalloc = Z_NULL;
     file->zlib.zfree = Z_NULL;
-    file->zlib.avail_in = zzip_file_header_csize (file->header);
-    file->zlib.next_in = zzip_file_header_to_data (file->header);
+    file->zlib.avail_in = zzip_file_header_csize (header);
+    file->zlib.next_in = zzip_file_header_to_data (header);
 
-    if (! zzip_file_header_data_deflated (file->header) ||
+    if (! zzip_file_header_data_deflated (header) ||
 	inflateInit2 (& file->zlib, -MAX_WBITS) != Z_OK)
     { free (file); return 0; }
 
-    return file;
+    return file; ____;
 }
 
 /** openening a file part wrapped within a (mmapped) zip archive
@@ -485,7 +495,7 @@ zzip_disk_entry_fopen (ZZIP_DISK* disk, ZZIP_DISK_ENTRY* entry)
  * the zip central directory with => zzip_disk_findfile and whatever
  * is found first is given to => zzip_disk_entry_fopen
  */
-ZZIP_DISK_FILE* _zzip_restrict
+zzip__new__ ZZIP_DISK_FILE*
 zzip_disk_fopen (ZZIP_DISK* disk, char* filename)
 {
     ZZIP_DISK_ENTRY* entry = zzip_disk_findfile (disk, filename, 0, 0);
